use crate::{commands::utils, Context, Error};
use cached::proc_macro::cached;
use chrono::{DateTime, Utc};
use log::{info, warn};
use reqwest::header::AUTHORIZATION;
use serde::Deserialize;

use super::Timeframe;

#[derive(Deserialize, Clone)]
pub struct CmoreEvent {
    #[serde(rename = "assetId")]
    id: u32,
    #[serde(rename = "asset")]
    data: CmoreData,
}

#[derive(Deserialize, Clone)]
pub struct CmoreData {
    #[serde(default)]
    description: String,
    title: String,
    subtitle: String,
    duration: i64,
    // #[serde(rename = "humanDuration")]
    // hduration: String,
    #[serde(rename = "liveBroadcastTime")]
    #[serde(with = "cmore_date")]
    start: DateTime<Utc>,
    // #[serde(rename = "mainCategoryTitle")]
    // category: String,
}

impl CmoreEvent {
    pub fn filter(&self, filter: &str) -> bool {
        if filter.is_empty() {
            return true;
        };
        if self.data.description.contains(filter) {
            return true;
        }
        if self.data.title.contains(filter) {
            return true;
        }
        if self.data.subtitle.contains(filter) {
            return true;
        }
        false
    }

    pub fn comp(&self, when: &Timeframe) -> bool {
        let now = Utc::now();
        match when {
            Timeframe::Everything => true,
            Timeframe::Current => {
                self.data.start <= now
                    && self.data.start + chrono::Duration::seconds(self.data.duration) >= now
            }
            Timeframe::Future => {
                self.data.start + chrono::Duration::seconds(self.data.duration) >= now
            }
            Timeframe::Past => {
                self.data.start + chrono::Duration::seconds(self.data.duration) <= now
            } // _ => self.data.times.end >= now,
        }
    }

    pub fn to_string(&self) -> String {
        format!(
            "```md\n[{title}]({subtitle})```(<t:{start}:R>-<t:{end}:R>) https://tom.al/ms/cm/{id}",
            title = self.data.title,
            subtitle = self.data.subtitle,
            start = self.data.start.timestamp(),
            end = self.data.start.timestamp() + self.data.duration,
            id = self.id
        )
    }

    pub fn get_key(&self) -> (DateTime<Utc>, String) {
        (self.data.start, self.data.description.clone())
    }
}

mod cmore_date {
    use chrono::{DateTime, TimeZone, Utc};
    use serde::{self, Deserialize, Deserializer};

    const FORMAT: &'static str = "%Y-%m-%dT%H:%M:%SZ";

    pub fn deserialize<'de, D>(deserializer: D) -> Result<DateTime<Utc>, D::Error>
    where
        D: Deserializer<'de>,
    {
        let s = String::deserialize(deserializer)?;
        Utc.datetime_from_str(&s, FORMAT)
            .map_err(serde::de::Error::custom)
    }
}

#[cached(time = 3600)]
pub async fn get_schedule() -> Option<Vec<CmoreEvent>> {
    let token = super::super::super::SETTINGS
        .read()
        .unwrap()
        .get_table("morningstreams")
        .unwrap()
        .get("token")
        .expect("Config error, please set the morningstreams[token] value")
        .clone()
        .into_string()
        .expect("Config error, please make sure morningstreams[token] is a string");
    let client = reqwest::Client::new();
    let req = client
        .get(format!(
            "https://api.morningstreams.com/api/hightier/cmore/schedule"
        ))
        .header(AUTHORIZATION, token)
        .send()
        .await;

    let result: Option<Vec<CmoreEvent>> = match req {
        Err(e) => {
            warn!("Error getting Cmore schedule {}", e);
            None
        }
        Ok(req) if req.status().as_u16() == 404 => {
            warn!("404 on getting cmore events");
            None
        }
        Ok(req) if req.status().as_u16() == 200 => {
            let data = req.json::<Vec<CmoreEvent>>().await;
            match data {
                Ok(d) => {
                    Some(d) // .iter().map(|e| e.asset.clone()).collect())
                }

                Err(e) => {
                    warn!("Error getting Cmore schedule {}", e);
                    None
                }
            }
        }
        Ok(req) => {
            warn!(
                "Unhandled status when parsing Cmore request {}",
                req.status()
            );
            None
        }
    };
    result
}

// Cmore events listing
#[poise::command(slash_command, ephemeral)]
pub async fn cmore(
    ctx: Context<'_>,
    #[description = "Filter sessions for when they are/were happening, defaults to future"]
    timeframe: Option<super::Timeframe>,
    #[description = "Content to filter on"] filter: Option<String>,
) -> Result<(), Error> {
    let tf = match timeframe {
        None => Timeframe::Future,
        Some(tf) => tf,
    };
    let events: Option<Vec<CmoreEvent>> = get_schedule().await;
    match events {
        None => {
            ctx.say("Unable to get the events, try again later (it's cached so wait a bit...)")
                .await?;
        }
        Some(evs) => {
            info!("Found {} events from cmore", evs.len());
            let filtered: Vec<String> = evs
                .into_iter()
                .filter(|e| e.comp(&tf))
                .filter(|e| match &filter {
                    None => true,
                    Some(f) => e.filter(f.as_str()),
                })
                .map(|e| e.to_string())
                .collect();
            let pages = utils::paginator(filtered, 1900, "\n".to_string());
            utils::paginate_string(ctx, pages).await?;
        }
    };
    Ok(())
}
