use config::{Config, File};
use glob::glob;
use lazy_static::lazy_static;
use log::{info, LevelFilter};
use poise::serenity_prelude as serenity;
use poise::serenity_prelude::UserId;
use regex::Regex;
use simplelog::SimpleLogger;
use std::{collections::HashSet, sync::RwLock};

mod helpers;
// mod entity;

// Types used by all command functions
type Error = Box<dyn std::error::Error + Send + Sync>;
type Context<'a> = poise::Context<'a, Data, Error>;

mod commands;

// Custom user data passed to all command functions
pub struct Data {
    database: mysql_async::Pool,
}

lazy_static! {
    static ref SETTINGS: RwLock<Config> = RwLock::new(Config::default());
}

/// Register application commands in this guild or globally
///
/// Run with no arguments to register in guild, run with argument "global" to register globally.
#[poise::command(
    prefix_command,
    hide_in_help,
    // required_permissions = "MANAGE_MESSAGES | ADMINISTRATOR",
    owners_only=true,
)]
async fn register(
    ctx: Context<'_>,
    #[flag] global: bool,
    #[flag] purge: bool,
    #[flag] not: bool,
    #[rest] rest: String,
) -> Result<(), Error> {
    // poise::builtins::register_application_commands(ctx, global).await?;

    let mut commands_builder = serenity::CreateApplicationCommands::default();
    let commands = &ctx.framework().options().commands;
    ctx.say(format!("Commands: {}", commands.len())).await?;
    let mut command_count: u8 = 0;
    for command in commands {
        if rest.eq_ignore_ascii_case("all") || (rest.contains(&command.name) ^ not) {
            command_count += 1;
            if let Some(slash_command) = command.create_as_slash_command() {
                commands_builder.add_application_command(slash_command);
            }
            if let Some(context_menu_command) = command.create_as_context_menu_command() {
                commands_builder.add_application_command(context_menu_command);
            }
        } else {
            info!("Not adding command {}", command.name);
        }
    }
    let commands_builder = serenity::json::Value::Array(commands_builder.0);

    let is_bot_owner = ctx.framework().options().owners.contains(&ctx.author().id);
    if global {
        if !is_bot_owner {
            ctx.say("Can only be used by bot owner").await?;
            return Ok(());
        }
        if purge {
            ctx.say(format!(
                "Purging global commands, must have been a mistake eh?"
            ))
            .await?;
            ctx.serenity_context()
                .http
                .create_global_application_commands(&serenity::json::Value::Null)
                .await?;
        } else {
            ctx.say(format!(
                "Registering {} commands out of {}...",
                command_count,
                commands.len()
            ))
            .await?;
            ctx.serenity_context()
                .http
                .create_global_application_commands(&commands_builder)
                .await?;
        }
    } else {
        let guild = match ctx.guild() {
            Some(x) => x,
            None => {
                ctx.say("Must be called in guild").await?;
                return Ok(());
            }
        };
        let is_guild_owner = ctx.author().id == guild.owner_id;

        if !is_guild_owner && !is_bot_owner {
            ctx.say("Can only be used by server owner").await?;
            return Ok(());
        }

        ctx.say(format!(
            "Registering {} commands out of {}...",
            command_count,
            commands.len()
        ))
        .await?;
        ctx.serenity_context()
            .http
            .create_guild_application_commands(guild.id.0, &commands_builder)
            .await?;
    }

    ctx.say("Done!").await?;

    Ok(())
}

/// Use the default configuration for setting up commands.
#[poise::command(
    prefix_command,
    hide_in_help,
    // required_permissions = "MANAGE_MESSAGES | ADMINISTRATOR",
    owners_only=true,
)]
async fn default_register(ctx: Context<'_>) -> Result<(), Error> {
    // poise::builtins::register_application_commands(ctx, global).await?;
    let is_bot_owner = ctx.framework().options().owners.contains(&ctx.author().id);
    let mut commands_builder_guild = serenity::CreateApplicationCommands::default();
    let commands = &ctx.framework().options().commands;
    ctx.say(format!("Commands: {}", commands.len())).await?;

    fn globals_filter(name: &String) -> bool {
        name.starts_with("invite") || name.starts_with("boop") || name.starts_with("fix")
    }

    // If the user is the bot owner, also correct the global commands.
    if is_bot_owner {
        let mut commands_builder_global = serenity::CreateApplicationCommands::default();
        // let global_commands = commands; //.iter().filter(|com| com.name.starts_with("invite")).collect();
        let mut global_count = 0;
        for command in commands {
            if globals_filter(&command.name) {
                global_count += 1;
                if let Some(slash_command) = command.create_as_slash_command() {
                    commands_builder_global.add_application_command(slash_command);
                }
                if let Some(context_menu_command) = command.create_as_context_menu_command() {
                    commands_builder_global.add_application_command(context_menu_command);
                }
            }
        }
        let commands_build_global = serenity::json::Value::Array(commands_builder_global.0);
        ctx.serenity_context()
            .http
            .create_global_application_commands(&commands_build_global)
            .await?;
        ctx.say(format!("Registered {} as global commands", global_count))
            .await?;
    }
    let mut command_count: u8 = 0;
    for command in commands {
        if !globals_filter(&command.name) {
            command_count += 1;
            if let Some(slash_command) = command.create_as_slash_command() {
                commands_builder_guild.add_application_command(slash_command);
            }
            if let Some(context_menu_command) = command.create_as_context_menu_command() {
                commands_builder_guild.add_application_command(context_menu_command);
            }
        }
    }
    let commands_builder = serenity::json::Value::Array(commands_builder_guild.0);
    let guild = match ctx.guild() {
        Some(x) => x,
        None => {
            return Ok(());
        }
    };
    let is_guild_owner = ctx.author().id == guild.owner_id;

    if !is_guild_owner && !is_bot_owner {
        ctx.say("Can only be used by server owner").await?;
        return Ok(());
    }
    ctx.say(format!(
        "Registering {} commands out of {}...",
        command_count,
        commands.len()
    ))
    .await?;
    ctx.serenity_context()
        .http
        .create_guild_application_commands(guild.id.0, &commands_builder)
        .await?;
    ctx.say("Done!").await?;
    Ok(())
}

async fn on_error(error: poise::FrameworkError<'_, Data, Error>) {
    // This is our custom error handler
    // They are many errors that can occur, so we only handle the ones we want to customize
    // and forward the rest to the default handler
    match error {
        poise::FrameworkError::Setup { error, .. } => panic!("Failed to start bot: {:?}", error),
        poise::FrameworkError::Command { error, ctx } => {
            println!("Error in command `{}`: {:?}", ctx.command().name, error,);
        }
        error => {
            if let Err(e) = poise::builtins::on_error(error).await {
                println!("Error while handling error: {}", e)
            }
        }
    }
}

/// Show this help menu
#[poise::command(prefix_command, track_edits, slash_command)]
async fn help(
    ctx: Context<'_>,
    #[description = "Specific command to show help about"]
    #[autocomplete = "poise::builtins::autocomplete_command"]
    command: Option<String>,
) -> Result<(), Error> {
    poise::builtins::help(
        ctx,
        command.as_deref(),
        poise::builtins::HelpConfiguration {
            extra_text_at_bottom: "\
This bot is very much a work in progress. Commands will change and may break.",
            show_context_menu_commands: true,
            ..Default::default()
        },
    )
    .await?;
    Ok(())
}

fn setup_config() -> Result<(), Error> {
    let mut setting = SETTINGS.write().unwrap();
    *setting = Config::builder()
        .add_source(
            glob("data/config/*")
                .unwrap()
                .map(|path| File::from(path.unwrap()))
                .collect::<Vec<_>>(),
        )
        .build()
        .unwrap();

    Ok(())
}

async fn app() -> Result<(), Error> {
    let commands = vec![
        help(),
        register(),
        default_register(),
        commands::invites::invites(),
        commands::schedule::schedule(),
        commands::roles::roles(),
        commands::boop(),
        commands::fix(),
        // commands::planning::get_command(),
        commands::get_links(),
        commands::coms::coms(),
    ];

    let discord = SETTINGS.read().unwrap().get_table("discord").unwrap();

    let mut owners: HashSet<UserId> = HashSet::new();
    owners.insert(UserId(117992484310745097));

    let prefix: String = discord
        .get("prefix")
        .expect("Config error, please set the discord[token] value")
        .to_owned()
        .into_string()
        .expect("Config error, please make sure discord[token] is a string");
    let options = poise::FrameworkOptions {
        commands: commands,
        prefix_options: poise::PrefixFrameworkOptions {
            prefix: Some(prefix.into()),
            edit_tracker: Some(poise::EditTracker::for_timespan(
                std::time::Duration::from_secs(3600),
            )),
            additional_prefixes: vec![
                poise::Prefix::Regex(Regex::new("(?i)ol'? rusty please").unwrap()),
                poise::Prefix::Regex(Regex::new("(?i)Oi Rusty please").unwrap()),
            ],
            ..Default::default()
        },
        on_error: |error| Box::pin(on_error(error)),
        pre_command: |ctx| {
            Box::pin(async move {
                println!("Executing command {}...", ctx.command().qualified_name);
            })
        },
        post_command: |ctx| {
            Box::pin(async move {
                println!("Executed command {}!", ctx.command().qualified_name);
            })
        },
        owners,
        ..Default::default()
    };

    // let discord = SETTINGS.read().unwrap().get_table("discord").unwrap();
    let token = discord
        .get("token")
        .expect("Config error, please set the discord[token] value")
        .clone()
        .into_string()
        .expect("Config error, please make sure discord[token] is a string");

    let db_settings = SETTINGS.read().unwrap().get_table("database").unwrap();

    // let db2: Result<HashMap<String, String>, Error> = db_settings.try_into();

    // let db = mysql_async::Pool::new(db_options);
    let db = helpers::setup_db(db_settings);

    let framework = poise::Framework::builder()
        .token(token)
        .setup(move |_ctx, _ready, _framework| {
            // Box::pin(async move { Ok(Data { database: database }) })
            Box::pin(async move { Ok(Data { database: db }) })
        })
        .options(options)
        .intents(
            serenity::GatewayIntents::non_privileged() | serenity::GatewayIntents::MESSAGE_CONTENT,
        );

    framework.run().await.unwrap();

    Ok(())
}

#[tokio::main]
async fn main() {
    print!("Starting the bot");
    setup_config().unwrap();
    // env_logger::init();
    // init_log();
    let log_default = LevelFilter::Warn;
    let logger = match SETTINGS
        .read()
        .unwrap()
        .get_table("logging")
        .unwrap_or_default()
        .get("level")
    {
        Some(val) => match val.clone().into_string() {
            Ok(level) => match level.to_lowercase().as_str() {
                "trace" => LevelFilter::Trace,
                "debug" => LevelFilter::Debug,
                "info" => LevelFilter::Info,
                "warn" => LevelFilter::Warn,
                "error" => LevelFilter::Error,
                _ => log_default,
            },
            _ => log_default,
        },
        _ => log_default,
    };

    println!("Logging level: {:?}", logger);

    SimpleLogger::init(logger, simplelog::Config::default()).unwrap();

    if let Err(e) = app().await {
        log::error!("{}", e);
        std::process::exit(1);
    }
}
